/****************************************/
/* MES/Copyleft Yukio Mituiwa,2005	*/
/*					*/
/*  2005/11/23 first release		*/
/*					*/
/****************************************/
#include "mes.h"
#include "coffdef.h"
#include <string.h>

typedef struct {
	char	*symbol;
	void	*func;
} mes_reloc;

static const mes_reloc symbols[] = {
	{"_stoi", stoi},
	{"_stoh", stoh},
	{"_task_sw_trap", task_sw_trap},
	{"_task_sw_ctl", task_sw_ctl},
	{"_wait_ms", wait_ms},
	{"_task_time", task_time},
	{"_usb_setup", usb_setup},
	{"_command_trap", command_trap},
	{"_usb_get_desc", usb_get_desc},
	{"_usb_get_conf", usb_get_conf},
	{"_usb_product", usb_product},
	{"_usb_alloc", usb_alloc},
	{"_usb_handle", usb_handle},
	{"_usb_status", usb_status},
	{"_usb_mount", usb_mount},
	{"_usb_write", usb_write},
	{"_usb_read", usb_read},
	{"_usb_allfree", usb_allfree},
	{"_get_dr", get_dr},
	{"_ddr_set", ddr_set},
	{"_ddr_reset", ddr_reset},
	{"_pe_set", pe_set},
	{"_pe_reset", pe_reset},
	{"_device_set_handler", set_handler},
	{"_device_malloc", __malloc},
	{"_device_free", __free},
	{"_device_open", __open},
	{"_device_close", __close},
	{"_device_write", __write},
	{"_device_read", __read},
	{"_device_seek", __seek},
	{"_device_ioctl", __ioctl},
	{0,0}
};

static void add(char* ptr, unsigned int address) {
	unsigned int	*long_ptr;

	long_ptr = (int*)ptr;
	*long_ptr += address;
}

int stack_size(char *coffarea) {
	unsigned short	secnum, optlen;
	int		c, ret;
	FILHDR		*filehdr;
	SCNHDR		*section;

	ret = 0x400;
	filehdr = (FILHDR*)coffarea;
	secnum = filehdr->f_nscns;
	optlen = filehdr->f_opthdr;
	section = (SCNHDR*)&(coffarea[FILHSZ + optlen]);
	for(c = 0;c < secnum;c++) {
		if(memcmp(section[c].s_name, ".stack", 6) == 0) {
			ret = section[c].s_vaddr;
			break;
		}
	}
	return ret;
}

int get_program_size(char *coffarea) {
	unsigned short magic, secnum, optlen, fileflag;
	unsigned int c, pgmsize, textsize, datasize, bsssize;
	FILHDR	 *filehdr;
	SCNHDR	 *section;

	textsize = 0;
	datasize = 0;
	bsssize = 0;
	filehdr = (FILHDR*)coffarea;
	secnum = filehdr->f_nscns;
	magic = filehdr->f_magic;
	optlen = filehdr->f_opthdr;
	fileflag = filehdr->f_flags;
	section = (SCNHDR*)&(coffarea[FILHSZ + optlen]);
	if((magic != H8MAGIC && magic != SH2MAGIC && magic != SH3MAGIC) || fileflag & F_RELFLG) return -1;
	for(c = 0;c < secnum;c++) {
		switch(section[c].s_flags) {
		case STYP_TEXT:
			textsize = section[c].s_size;
			break;
		case STYP_DATA:
			datasize = section[c].s_size;
			break;
		case STYP_BSS:
			bsssize = section[c].s_size;
			break;
		}
	}
	pgmsize = textsize + datasize + bsssize;
	return pgmsize;
}

int coff2bin(char* binarea, char* coffarea) {
	unsigned short	magic, secnum, optlen, fileflag;
	unsigned short	t_relnum, d_relnum, relnum, relscnum;
	unsigned short	relflag, value_w;
	unsigned int	i, c, s, n, symptr, symnum;
	unsigned int	textsize, datasize, bsssize, pgmsize;
	unsigned int	textadr, dataadr, bssadr;
	unsigned int	textptr, dataptr, bssptr;
	unsigned int	t_relptr, d_relptr, relptr;
	unsigned int	relindex, reladdr, reloff, addend, sec_off;
	int		ret, symbolp;
	char		*symbolname;
	FILHDR		*filehdr;
	SCNHDR		*section;
	RELOC		*relinfo;
	SYMENT		*syminfo;

	ret = 0;
	t_relnum = 0;
	d_relnum = 0;
	textsize = 0;
	datasize = 0;
	bsssize = 0;
	textadr = 0;
	dataadr = 0;
	textptr = 0;
	dataptr = 0;
	t_relptr = 0;
	d_relptr = 0;
	filehdr = (FILHDR*)coffarea;
	magic = filehdr->f_magic;
	secnum = filehdr->f_nscns;
	symptr = filehdr->f_symptr;
	symnum = filehdr->f_nsyms;
	optlen = filehdr->f_opthdr;
	fileflag = filehdr->f_flags;
	section = (SCNHDR*)&(coffarea[FILHSZ + optlen]);
#if defined(__H8300H__)
	if(magic != H8MAGIC || fileflag & F_RELFLG) return -1;
#endif
#if defined(__sh2__)
	if(magic != SH2MAGIC || fileflag & F_RELFLG) return -1;
#endif
#if defined(__sh3__)
	if(magic != SH3MAGIC || fileflag & F_RELFLG) return -1;
#endif
	for(c = 0;c < secnum;c++) {
		switch(section[c].s_flags) {
		case STYP_TEXT:
			textadr = section[c].s_vaddr;
			textsize = section[c].s_size;
			textptr = section[c].s_scnptr;
			t_relptr = section[c].s_relptr;
			t_relnum = section[c].s_nreloc;
			break;
		case STYP_DATA:
			dataadr = section[c].s_vaddr;
			datasize = section[c].s_size;
			dataptr = section[c].s_scnptr;
			d_relptr = section[c].s_relptr;
			d_relnum = section[c].s_nreloc;
			break;
		case STYP_BSS:
			bssadr = section[c].s_vaddr;
			bsssize = section[c].s_size;
			bssptr = section[c].s_scnptr;
			break;
		}
	}
	pgmsize = textsize + datasize + bsssize;
	bzero(binarea, pgmsize);
	memcpy(&(binarea[textadr]), &(coffarea[textptr]), textsize);
	memcpy(&(binarea[dataadr]), &(coffarea[dataptr]), datasize);
	for(s = 0;s <= 1;s++) {
		if(s == 0) {
			relptr = t_relptr;
			relnum = t_relnum;
			sec_off = textadr;
		} else if(s == 1) {
			relptr = d_relptr;
			relnum = d_relnum;
			sec_off = dataadr;
		} else {
			break;
		}
		relinfo = (RELOC*)&(coffarea[relptr]);
		syminfo = (SYMENT*)&(coffarea[symptr]);
		for(c = 0;c < relnum;c++) {
			reladdr = relinfo[c].r_vaddr;
			relindex = relinfo[c].r_symndx;
			relflag = relinfo[c].r_type;
			addend = relinfo[c].r_offset;
			if(relindex >= 0 && relindex < symnum) {
				reloff = syminfo[relindex].e_value;
				relscnum = syminfo[relindex].e_scnum;
				if(relscnum == 0) {
					if(syminfo[relindex].e.e.e_zeroes == 0) {
						symbolp = symnum * sizeof(SYMENT) + symptr + syminfo[relindex].e.e.e_offset;
						symbolname = &(coffarea[symbolp]);
					} else {
						symbolname = syminfo[relindex].e.e_name;
					}
					n = strlen(symbolname);
					for(i = 0;symbols[i].symbol != 0;i++) {
						if(strcmp(symbolname, symbols[i].symbol) == 0) {
							add(&(binarea[reladdr]), (int)symbols[i].func);
							break;
						}
					}
					if(symbols[i].symbol == 0) return -1;
				} else {
					if(magic == SH2MAGIC || magic == SH3MAGIC) {
						add(&(binarea[reladdr]), (int)binarea);
					} else if(magic == H8MAGIC) {
						switch(relflag) {
						case R_PCRBYTE:
							binarea[reladdr] = (reloff + addend) - reladdr;
							break;
						case R_PCRWORD:
							value_w = (reloff + addend) - reladdr - 1;
							binarea[reladdr] = value_w >> 8;
							binarea[reladdr + 1] = value_w;
							break;
/*						case R_PCRLONG:
							value_l = (reloff + addend)  - reladdr;
							binarea[reladdr] = value_l >> 24;
							binarea[reladdr + 1] = value_l >> 16;
							binarea[reladdr + 2] = value_l >> 8;
							binarea[reladdr + 3] = value_l;
							break;	*/
						case R_MOV24B1:
						case R_MOVL1:
						case R_RELLONG:
						case R_JMPL1:
							add(&(binarea[reladdr]), (int)binarea + reloff + addend);
							break;
						default:
							return -1;
				    		}
					}
				}
			}
		}
	}
	return 0;
}
