/****************************************/
/* MES/Copyleft Yukio Mituiwa,2005	*/
/*					*/
/*  2005/2/14 first release		*/
/*					*/
/****************************************/
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include "mes.h"
#include "id.h"

#define	STDFILENUM	2
#define MAXDEV		32
#define MAXMINOR	16
typedef struct {
	unsigned int	id[MAXMINOR], dev_id;
	int		time[MAXMINOR], remain[MAXMINOR];
	Functions	*ptr;
	char		name[8];
} Devices;
static Devices	*dev;

extern void* device_entries[];

static int get_number(char *name, int *major, int *minor) {
	int	i, n;
	char	*p, buf[8];

	n = strlen(name);
	if(n < 2) return -1;
	if(!isdigit(name[n - 1])) return -1;
	if(!isalpha(name[0])) return -1;
	for(i = n - 1;i > 0;i--) {
		if(!isdigit(name[i])) break;
	}
	if(i > 6) return -1;
	p = &name[i + 1];
	*minor = atoi(p);
	if(*minor >= MAXMINOR) return -1;
	memcpy(buf, name, 8);
	buf[i + 1] = 0;
	for(i = 0;i < MAXDEV;i++) {
		if(strcmp(dev[i].name, buf) == 0) {
			*major = i;
			return 0;
		}
	}
	return -1;		
}

static int fd2number(int fd, int *major, int *minor) {
	if(!(fd & DEVFD_MASK)) return -1;
	*major = (fd >> 8) & 0xff;
	*minor = fd & 0xff;
	return 0;
}

void device_init(void) {
	int		i, j;
	unsigned int	addr;
	Functions*	(*init_func)();

	dev = (Devices *)__malloc(SuperTask, sizeof(Devices) * MAXDEV);
	for(i = 0;i < MAXDEV;i++) {
		for(j = 0;j < MAXMINOR;j++) dev[i].id[j] = 0;
		dev[i].ptr = 0;
		dev[i].name[0] = 0;
		dev[i].dev_id = 0;
	}
	for(i = 0;i < MAXDEV;i++) {
		addr = (int)device_entries[i];
		if(addr == 0xffffffff) break;
		if(addr > 0) {
			init_func = device_entries[i];
			dev[i].ptr = (*init_func)();
			strcpy(dev[i].name, dev[i].ptr->name);
		}
	}
}

static int add_device_common(int id, void *func, int flag) {
	int		i;

	for(i = 0;i < MAXDEV;i++) {
		if(dev[i].ptr == 0) {
			dev[i].ptr = func;
			dev[i].dev_id = id;
			strcpy(dev[i].name, (dev[i].ptr)->name);
			if(flag == 0) {
				set_task_stay(id);
				task_wakeup(id, 0);
			}
			return 0;
		}
	}
	return -1;
}

int add_device(int id, void *func) {
	return add_device_common(id, func, 0);
}

int add_device_multi(int id, void *func) {
	return add_device_common(id, func, 1);
}

int del_device(int id) {
	int		i, j;

	for(i = 0;i < MAXDEV;i++) {
		if(dev[i].ptr == 0) continue;
		if(dev[i].dev_id == id) {
			for(j = 0;j < MAXMINOR;j++) {
				if(dev[i].id[j] != 0) return -1;
			}
			dev[i].ptr = 0;
			dev[i].dev_id = 0;
			return 0;
		}
	}
	return -1;
}

int __open(int id, char *name, int option){
	unsigned int	major, minor, fd;
	int		ret;

	if(get_number(name, &major, &minor) != 0) {
		ret = OpenFile(id, name, option);
		if(ret == -1)  {
			fd = -1;
		} else {
			fd = ret | FATFD_MASK;
		}
		return fd;
	}
	if(dev[major].id[minor] != 0) return -1;
	if(dev[major].ptr == 0) return -1;
	if(dev[major].name[0] == 0) return -1;
	if((*(dev[major].ptr->open_dev))(minor, option) != 0) return -1;
	fd = (major << 8) + minor + DEVFD_MASK;
	dev[major].id[minor] = id;
	dev[major].time[minor] = -1;
	dev[major].remain[minor] = -1;
	return fd;
}

int __close(int id, int fd){
	unsigned int	major, minor;

	if(fd2number(fd, &major, &minor) != 0) {
		return CloseFile(id, fd & FAT_MASK);
	}
	if(dev[major].id[minor] != id) return -1;
	if((*(dev[major].ptr->close_dev))(minor) != 0) return -1;
	dev[major].id[minor] = 0;
	return 0;
}

int __write(int id, int fd, char *data, int size){
	unsigned int	major, minor, _fd, _id;
	int		ret;

	if(fd <= STDFILENUM) {
		if(fd == 2) {
			_fd = get_stderr(id);
		} else {
			_fd = get_stdio(id);
		}
		if(fd2number(_fd, &major, &minor) == 0) {
			ret = (*(dev[major].ptr->write_dev))(minor, data, size);
		} else {
			_id = GetFileId(_fd);
			ret = WriteFile(_id, _fd & FAT_MASK, data, size);
		}
		return ret;
	}
	if(fd2number(fd, &major, &minor) != 0) {
		ret = WriteFile(id, fd & FAT_MASK, data, size);
	} else if(dev[major].id[minor] == id) {
		if(dev[major].id[minor] != id) return -1;
		ret = (*(dev[major].ptr->write_dev))(minor, data, size);
	} else {
		ret = -1;
	}
	return ret;
}

int __read(int id, int fd, char *data, int size){
	unsigned int	major, minor, _fd, _id;
	int		ret, flag;

	flag = 0;
	if(fd <= STDFILENUM) {
		if(fd == 2) {
			_fd = get_stderr(id);
		} else {
			_fd = get_stdio(id);
		}
		if(fd2number(_fd, &major, &minor) == 0) {
			flag = 1;
		} else {
			_id = GetFileId(_fd);
			ret = ReadFile(_id, _fd & FAT_MASK, data, size);
			return ret;
		}
	} else {
		if(fd2number(fd, &major, &minor) != 0) {
			ret = ReadFile(id, fd & FAT_MASK, data, size);
		} else if(dev[major].id[minor] == id) {
			if(dev[major].id[minor] != id) return -1;
			flag = 1;
		} else {
			ret = -1;
		}
	}
	if(flag == 1) {
		ret = (*(dev[major].ptr->read_dev))(minor, data, size);
		if(dev[major].time[minor] != -1) {
			if(ret > 0) {
				dev[major].remain[minor] = -1;
			} else {
				ret = -2;
				if(dev[major].remain[minor] == -1 || dev[major].time[minor] == 0) {
					dev[major].remain[minor] = task_time() + dev[major].time[minor];
				} else {
					if(dev[major].remain[minor] < task_time()) {
						dev[major].remain[minor] = -1;
						ret = 0;
					}
				}
			}
		}
	}
	return ret;
}

int __seek(int id, int fd, int position) {
	unsigned int	major, minor, _fd, _id;
	int		ret;

	if(fd <= STDFILENUM) {
		if(fd == 2) {
			_fd = get_stderr(id);
		} else {
			_fd = get_stdio(id);
		}
		if(fd2number(_fd, &major, &minor) == 0) {
			ret = (*(dev[major].ptr->seek_dev))(minor, position);
		} else {
			_id = GetFileId(_fd);
			ret = SeekFile(_id, _fd & FAT_MASK, position);
		}
		return ret;
	}
	if(fd2number(fd, &major, &minor) != 0) {
		ret = SeekFile(id, fd & FAT_MASK, position);
	} else if(dev[major].id[minor] == id) {
		if(dev[major].id[minor] != id) return -1;
		ret = (*(dev[major].ptr->seek_dev))(minor, position);
	} else {
		ret = -1;
	}
	return ret;
}

int __ioctl(int id, int fd, int data, int op) {
	unsigned int	major, minor, _fd;
	int		ret;

	if(fd <= STDFILENUM) {
		if(fd == 2) {
			_fd = get_stderr(id);
		} else {
			_fd = get_stdio(id);
		}
	} else {
		_fd = fd;
	}
	if(fd2number(_fd, &major, &minor) != 0) return -1;
	if(fd > STDFILENUM && dev[major].id[minor] != id) return -1;
	if(op == DEV_BLOCK) {
#if defined(__H8300H__) || defined(__sh2__)
		dev[major].time[minor] = (data / 3) * 2;
#endif
#if defined(__sh3__)
		dev[major].time[minor] = data << 1;
#endif
		dev[major].remain[minor] = -1;
		ret = 0;
	} else {
		ret = (*(dev[major].ptr->ioctl_dev))(minor, data, op);
	}
	return ret;
}

int __flush(int id, int fd) {
	return FlushFile(id, fd & FAT_MASK);
}

void device_interval() {
	int	major;

	for(major = 0;major < MAXDEV;major++) {
		if(dev[major].ptr != 0) {
			if(dev[major].ptr->poll_dev != 0) {
				(*(dev[major].ptr->poll_dev))();
			}
		}
	}
}

void free_device(int id) {
	int	major, minor;

	for(major = 0;major < MAXDEV;major++) {
		for(minor = 0;minor < MAXMINOR;minor++) {
			if(dev[major].id[minor] == id) {
				(*(dev[major].ptr->close_dev))(minor);
				dev[major].id[minor] = 0;
			}
		}
	}
}
