/* ELF Header (32-bit implementations) */

typedef struct {
  unsigned char	e_ident[16];		/* ELF "magic number" */
  unsigned short e_type;		/* Identifies object file type */
  unsigned short e_machine;		/* Specifies required architecture */
  unsigned int e_version;		/* Identifies object file version */
  unsigned int e_entry;			/* Entry point virtual address */
  unsigned int e_phoff;			/* Program header table file offset */
  unsigned int e_shoff;			/* Section header table file offset */
  unsigned int e_flags;			/* Processor-specific flags */
  unsigned short e_ehsize;		/* ELF header size in bytes */
  unsigned short e_phentsize;		/* Program header table entry size */
  unsigned short e_phnum;		/* Program header table entry count */
  unsigned short e_shentsize;		/* Section header table entry size */
  unsigned short e_shnum;		/* Section header table entry count */
  unsigned short e_shstrndx;		/* Section header string table index */
} __attribute__ ((packed)) ElfHdr;

/* Program header */

typedef struct {
  unsigned int	p_type;			/* Identifies program segment type */
  unsigned int	p_offset;		/* Segment file offset */
  unsigned int	p_vaddr;		/* Segment virtual address */
  unsigned int	p_paddr;		/* Segment physical address */
  unsigned int	p_filesz;		/* Segment size in file */
  unsigned int	p_memsz;		/* Segment size in memory */
  unsigned int	p_flags;		/* Segment flags */
  unsigned int	p_align;		/* Segment alignment, file & memory */
} __attribute__ ((packed)) ElfPhdr;

/* Section header */

typedef struct {
  unsigned int	sh_name;		/* Section name, index in string tbl */
  unsigned int	sh_type;		/* Type of section */
  unsigned int	sh_flags;		/* Miscellaneous section attributes */
  unsigned int	sh_addr;		/* Section virtual addr at execution */
  unsigned int	sh_offset;		/* Section file offset */
  unsigned int	sh_size;		/* Size of section in bytes */
  unsigned int	sh_link;		/* Index of another section */
  unsigned int	sh_info;		/* Additional section information */
  unsigned int	sh_addralign;		/* Section alignment */
  unsigned int	sh_entsize;		/* Entry size if section holds table */
} __attribute__ ((packed)) ElfShdr;

/* Symbol table entry */

typedef struct {
  unsigned int	st_name;		/* Symbol name, index in string tbl */
  unsigned int	st_value;		/* Value of the symbol */
  unsigned int	st_size;		/* Associated symbol size */
  unsigned char	st_info;		/* Type and binding attributes */
  unsigned char	st_other;		/* No defined meaning, 0 */
  unsigned short st_shndx;		/* Associated section index */
} __attribute__ ((packed)) ElfSym;

typedef struct {
  unsigned int est_shndx;		/* Section index */
} __attribute__ ((packed)) ElfSymShndx;

/* Note segments */

typedef struct {
  unsigned int	namesz;			/* Size of entry's owner string */
  unsigned int	descsz;			/* Size of the note descriptor */
  unsigned int	type;			/* Interpretation of the descriptor */
  char		name;			/* Start of the name+desc data */
} __attribute__ ((packed)) ElfNote;

/* Relocation Entries */
typedef struct {
  unsigned int r_offset;		/* Location at which to apply the action */
  unsigned int r_info;			/* index and type of relocation */
} __attribute__ ((packed)) ElfRel;

typedef struct {
  unsigned int r_offset;		/* Location at which to apply the action */
  unsigned int r_info;			/* index and type of relocation */
  unsigned int r_addend;		/* Constant addend used to compute value */
} __attribute__ ((packed)) ElfRela;

/* dynamic section structure */

typedef struct {
  unsigned int	d_tag;			/* entry tag value */
  union {
    unsigned int d_val;
    unsigned int d_ptr;
  } d_un;
} __attribute__ ((packed)) ElfDyn;

/* This structure appears in a SHT_GNU_verdef section.  */

typedef struct {
  unsigned short vd_version;
  unsigned short vd_flags;
  unsigned short vd_ndx;
  unsigned short vd_cnt;
  unsigned int vd_hash;
  unsigned int vd_aux;
  unsigned int vd_next;
} __attribute__ ((packed)) ElfVerdef;

/* This structure appears in a SHT_GNU_verdef section.  */

typedef struct {
  unsigned int vda_name;
  unsigned int vda_next;
} __attribute__ ((packed)) ElfVerdaux;

/* This structure appears in a SHT_GNU_verneed section.  */

typedef struct {
  unsigned short vn_version;
  unsigned short vn_cnt;
  unsigned int vn_file;
  unsigned int vn_aux;
  unsigned int vn_next;
} __attribute__ ((packed)) ElfVerneed;

/* This structure appears in a SHT_GNU_verneed section.  */

typedef struct {
  unsigned int vna_hash;
  unsigned short vna_flags;
  unsigned short vna_other;
  unsigned int vna_name;
  unsigned int vna_next;
} ElfVernaux;

/* This structure appears in a SHT_GNU_versym section.  This is not a
   standard ELF structure; ELF just uses Elf32_Half.  */

typedef struct {
  unsigned short vs_vers;
} __attribute__ ((packed)) ElfVersym;

/* Structure for syminfo section.  */
typedef struct
{
  unsigned short si_boundto;
  unsigned short si_flags;
} __attribute__ ((packed)) ElfSyminfo;


/* This structure appears on the stack and in NT_AUXV core file notes.  */
typedef struct
{
  unsigned int a_type;
  unsigned int a_val;
} __attribute__ ((packed)) ElfAuxv;
