/****************************************/
/* MES/Copyleft Yukio Mituiwa,2005	*/
/*					*/
/*  2005/2/14 first release		*/
/*  2005/9/15 irq task delete		*/
/*					*/
/****************************************/
#include "mes.h"
#include <regdef.h>
#include <string.h>
#if defined(__arm__)
  #define VECNUM 32
  static int	int_bits;
  static int	int_req;
#endif
#if defined(__H8300H__)
  #define VECNUM 64
  extern volatile int ramvectors;
  extern volatile int segment;
#endif
#if defined(__sh__)
  #define VECNUM 176
  extern volatile int ramvectors;
#endif
void	dummy_int(void);

void int_enable(void) {
#if defined(__arm__)
	int	cpsr;

	asm("mrs %0,cpsr" : "=r" (cpsr));
	cpsr &= ~I_BIT;
	asm("msr cpsr_c,%0" :: "r" (cpsr));
#endif
#if defined(__H8300H__)
	asm("andc #0x3f, ccr");
#endif
#if defined(__sh2__)
	int	sr;

	asm("stc sr,%0" : "=r" (sr));
	sr &= ~0xf0;
	asm("ldc %0,sr" :: "r" (sr));
#endif
#if defined(__sh3__)
	int	sr;

	asm("stc sr,%0" : "=r" (sr));
	sr &= ~0x100000f0;
	asm("ldc %0,sr" :: "r" (sr));
#endif
}

int int_disable(void) {
	int	reg;
#if defined(__arm__)
	asm("mrs %0,cpsr" : "=r" (reg));
	reg |= I_BIT;
	asm("msr cpsr_c,%0" :: "r" (reg));
#endif
#if defined(__H8300H__)
	asm("orc #0xc0, ccr");
#endif
#if defined(__sh__)
	asm("stc sr,%0" : "=r" (reg));
	reg |= 0xf0;
	asm("ldc %0,sr" :: "r" (reg));
#endif
	return reg;
}

int set_handler(int id, unsigned int number, void (*handler)()) {
#if defined(__arm__)
	if(number >= VECNUM) return -1;

	if((number == US0_ID) || (number == US1_ID)) {
		(AIC_BASE->AIC_SMR[number]) = AIC_SRCTYPE_EXT_NEGATIVE_EDGE + 7;
	} else {
		(AIC_BASE->AIC_SMR[number]) = AIC_SRCTYPE_EXT_NEGATIVE_EDGE;
	}
	(AIC_BASE->AIC_SVR[number]) = (at91_reg)handler;
	AIC_ICCR = 1 << number;
	AIC_IECR = 1 << number;
	int_bits |= 1 << number;
#endif
#if defined(__H8300H__)
	unsigned int	inst, got, ret;
	volatile int	*vectors, *gotsave;

	if(number < 1 || number >= VECNUM) return -1;
	inst = (int)handler;
	inst += 0x5a000000;
	vectors = &ramvectors;
	vectors[number] = (volatile int)inst;
	gotsave = &segment;
	gotsave[number] = 0;
	ret = get_reg(id, 5, &got);
	if(ret == 0) gotsave[number] = got;
	return 0;
#endif
#if defined(__sh2__)
	volatile int	*vectors;

	if(number < 1 || number >= VECNUM) return -1;
	vectors = &ramvectors;
	vectors[number] = (volatile int)handler;
	return 0;
#endif
#if defined(__sh3__)
	volatile int	*vectors;

	if(number >= 0xb00) return -1;
	vectors = &ramvectors;
	vectors[number >> 5] = (volatile int)handler;
	return 0;
#endif
}

#if defined(__arm__)
void irq_func(void) {
	void	(*func)();
	int	isr;

	func = (void*)AIC_IVR;
	isr = AIC_ISR;
	switch(isr) {
	case US0_ID:
	case US1_ID:
	case TC0_ID:
		(*func)();
		break;
	default:
		int_req |= 1 << isr;
	}
	AIC_ICCR = 1 << isr;
	AIC_EOICR = 1;
}

void core_func(int pc) {
	volatile int w;

	AIC_IDCR = 0xffffffff;
	AIC_IECR = 1 << US0_ID;
	int_enable();
	for(w = 0;w < 500000;w++);
	write_hex((char*)&pc, 4);
	__write(SuperTask, 0, " BUS Error!", 10);
	for(;;);
}
#endif

void interrupt_init(void) {
#if defined(__arm__)
	int_bits = 0;
	int_req = 0;
#endif
#if defined(__sh2__)
	int	i;
	volatile int	*vectors;

	asm("ldc %0,vbr" :: "r" (&ramvectors));
	memcpy((char*)&ramvectors, (char*)0, 16);
	vectors = &ramvectors;
	for(i = 4;i < VECNUM;i++)vectors[i] = (volatile int)dummy_int;
#endif
}
