/****************************************/
/* MES/Copyleft Yukio Mituiwa,2005	*/
/*					*/
/*  2006/6/16 first release		*/
/*					*/
/****************************************/
#include <macro.h>
#include <regdef.h>
#include <module.h>
#include "ethernet.h"
#include "ne.h"

#ifdef _KERNEL_
  #include "mes.h"
  #include "conf.h"
  #define device_main(arg1,arg2) init_ne(void)
  extern BUSConf ne_config[];
#endif

#define NE_NUM	2
static NEInfo *ne;

static void ne_outb(int minor, int port, char data) {
	volatile char *reg;

	reg = ne[minor].base;
	reg[(ne[minor].buswidth == 16) ? port * 2 : port] = data;
}

static char ne_inb(int minor, int port) {
	char	data;
	volatile char *reg;

	reg = ne[minor].base;
	data = reg[(ne[minor].buswidth == 16) ? port * 2 : port];
	return data;
}

static void ne_outw(int minor, int port, unsigned short data) {
	unsigned int	value;

	value = data & 0xff;
	ne_outb(minor, port, value);
	value = (data >> 8) & 0xff;
	ne_outb(minor, port + 1, value);
}

static void ne_reset(int minor) {
	volatile int	w;
	char		a;

	task_sw_ctl(1);
	a = ne_inb(minor, NE_ASIC_RESET);
	ne_outb(minor, NE_ASIC_RESET, a);
	a = ne_inb(minor, NE_ASIC_RESET2);
	for(w = 0;w < 0x40000;w++);
	ne_outb(minor, NE_P0_COMMAND, NE_CR_RD2 | NE_CR_STP);
	for(w = 0;w < 0x40000;w++);
	task_sw_ctl(0);
}

static int ne_writemem(int minor, char *src, unsigned int dest, unsigned int size) {
	int	i, packetsize;
	char	f, v;

	if(size == 0) return 0;
	packetsize = size;
	if(packetsize < ETHER_MIN_LEN) packetsize = ETHER_MIN_LEN;
	task_sw_ctl(1);
	ne_outb(minor, NE_P0_COMMAND, NE_CR_RD2 | NE_CR_STA);
	ne_outb(minor, NE_P0_ISR, 0);
	ne_outw(minor, NE_P0_RBCR0, packetsize);
	ne_outw(minor, NE_P0_RSAR0, dest);
	ne_outb(minor, NE_P0_COMMAND, NE_CR_RD1 | NE_CR_STA);
	for(i = 0;i < size;i++) ne_outb(minor, NE_ASIC_DATA, src[i]);
	for(;i < ETHER_MIN_LEN;i++) ne_outb(minor, NE_ASIC_DATA, 0);
	task_sw_ctl(0);
	i = 0;
	while(!(ne_inb(minor, NE_P0_ISR) & NE_ISR_RDC)) {
		if(++i > 8000000) return -1;
	}
	return 0;
}

static void ne_readmem(int minor, unsigned int src, char *dest, unsigned int size) {
	int		i;

	if(size == 0) return;
	task_sw_ctl(1);
	ne_outb(minor, NE_P0_COMMAND, NE_CR_RD2 | NE_CR_STA);
	ne_outw(minor, NE_P0_RBCR0, size);
	ne_outw(minor, NE_P0_RSAR0, src);
	ne_outb(minor, NE_P0_COMMAND, NE_CR_RD0 | NE_CR_STA);
	for(i = 0;i < size;i++) dest[i] = ne_inb(minor, NE_ASIC_DATA);
	task_sw_ctl(0);
}

static void ne_mem_probe(int minor) {
	unsigned int	a, flag;
	char		buffer[20], message[20] = "NE2000 test message";

	ne[minor].start = 0;
	ne[minor].size = 0;
	bzero(buffer, 20);
	for(a = 0x100;a < 0x10000;a += 0x100) {
		if(ne_writemem(minor, buffer, a, 20) == -1) return;
	}
	flag = 0;
	for(a = 0x100;a < 0x10000;a += 0x100) {
		ne_readmem(minor, a, buffer, 20);
		if(memcmp(buffer, message, 20) == 0) continue;
		ne_writemem(minor, message, a, 20);
		ne_readmem(minor, a, buffer, 20);
		if(memcmp(buffer, message, 20) == 0) {
			if(flag == 0) {
				flag = 1;
				ne[minor].start = a;
				ne[minor].size = 0x100;
			} else {
				ne[minor].size += 0x100;
			}
		} else {
			flag = 0;
		}
	}
}

static void ne_stop(int minor) {
	ne_outb(minor, NE_P0_COMMAND, NE_CR_STP);
	ne_outw(minor, NE_P0_RBCR0, 0);
	ne_outb(minor, NE_P0_RCR, NE_RCR_MON);
	ne_outb(minor, NE_P0_TCR, NE_TCR_LB0);
}

static void ne_start(int minor) {
	ne_outb(minor, NE_P0_COMMAND, NE_CR_STA);
	ne_outb(minor, NE_P0_RCR, NE_RCR_AB);
	ne_outb(minor, NE_P0_TCR, 0);
}

static int ne_probe(int minor) {
	int	i;
	char	eeprom[32], cmp, n;

	ne_reset(minor);
	ne_outb(minor, NE_P0_PSTART, 0x5a);
	ne_outb(minor, NE_P0_COMMAND, NE_CR_PS1 | NE_CR_STP);
	for(n = 1;n != 0;n <<= 1) {
		ne_outb(minor, NE_P1_PAR0, n);
		cmp = ne_inb(minor, NE_P1_PAR0);
		if(cmp != n) return -1;
	}
	ne_outb(minor, NE_P0_COMMAND, NE_CR_PS2 | NE_CR_STP);
	cmp = ne_inb(minor, NE_P0_PSTART);
	ne_outb(minor, NE_P0_COMMAND, NE_CR_PS0 | NE_CR_STP);
	if(cmp == 0x5a) {
		ne[minor].nic = NIC_DP8390;
	} else {
		ne_outb(minor, NE_P0_PSTART, 0x5a);
		cmp = ne_inb(minor, NE_P0_PSTART);
		if(cmp == 0x5a) {
			ne[minor].nic = NIC_AX88796B;
		} else {
			return -1;
		}
	}
	ne_outb(minor, NE_P0_RCR, NE_RCR_MON);
	ne_outb(minor, NE_P0_DCR, NE_DCR_BOS | NE_DCR_FT1 | NE_DCR_LS);
	ne_outb(minor, NE_P0_TCR, NE_TCR_LB0);
	ne_mem_probe(minor);
	ne_readmem(minor, 0, eeprom, 32);
	for(i = 0;i < 6;i++) ne[minor].mac[i] = eeprom[i * 2];
	ne_outb(minor, NE_P0_ISR, 0);
	return 0;
}

static int ne_transmit(int minor, int size) {
	int	packetsize;

	while(ne_inb(minor, NE_P0_COMMAND) & NE_CR_TXP);
	packetsize = size;
	if(packetsize < ETHER_MIN_LEN) packetsize = ETHER_MIN_LEN;
	ne_outb(minor, NE_P0_COMMAND, NE_CR_PS0 | NE_CR_RD2 | NE_CR_STA);
	ne_outw(minor, NE_P0_TBCR0, packetsize);
	ne_outb(minor, NE_P0_COMMAND, NE_CR_PS0 | NE_CR_TXP | NE_CR_RD2 | NE_CR_STA);
	while(ne_inb(minor, NE_P0_COMMAND) & NE_CR_TXP);
	return (ne_inb(minor, NE_P0_TSR) & NE_TSR_PTX) ? 0 : -1;
}

unsigned int ne_get_config(int minor) {
	unsigned int	i, data;

	ne_outb(minor, NE_P0_COMMAND, NE_CR_PS3);
	data = 0;
	for(i = 0;i < 4;i++) {
		data <<= 8;
		data |= (unsigned int)ne_inb(minor, NE_P3_CONFIG0 + i) & 0xff;
	}
	ne_outb(minor, NE_P0_COMMAND, NE_CR_PS0);
	return data;
}

static void ne_ring_init(int minor) {
	int	page;

	page = ne[minor].start >> 8;
	ne_outb(minor, NE_P0_TPSR, page);
	page = ne[minor].rx_start >> 8;
	ne_outb(minor, NE_P0_BNRY, page);
	ne_outb(minor, NE_P0_PSTART, page);
	page = ne[minor].rx_end >> 8;
	ne_outb(minor, NE_P0_PSTOP, page);
	ne_outb(minor, NE_P0_COMMAND, NE_CR_PS1 | NE_CR_STP);
	page = ne[minor].rx_start >> 8;
	page++;
	ne_outb(minor, NE_P1_CURR, page);
	ne_outb(minor, NE_P0_COMMAND, NE_CR_STP);
}

static void ne_setup(int minor) {
	int	m;

	ne[minor].rx_start = ne[minor].start + NE_TX_PAGE_SIZE;
	ne[minor].rx_end = ne[minor].start + ne[minor].size;
	ne_outb(minor, NE_P0_COMMAND, NE_CR_STP);
	ne_ring_init(minor);
	ne_outb(minor, NE_P0_ISR, 0xff);
	ne_outb(minor, NE_P0_IMR, NE_IMR_PRXE | NE_IMR_RXEE | NE_IMR_TXEE | NE_IMR_OVWE | NE_IMR_CNTE);
	ne_outb(minor, NE_P0_COMMAND, NE_CR_PS1 | NE_CR_STP);
	for(m = 0;m < 6;m++) ne_outb(minor, NE_P1_PAR0 + m, ne[minor].mac[m]);
	for(m = 0;m < 8;m++) ne_outb(minor, NE_P1_MAR0 + m, 0);
	ne_outb(minor, NE_P0_COMMAND, NE_CR_STP);
}

static int ne_attach(int minor) {
	int	ret;

	ret = ne_probe(minor);
	if(ret == 0) ne_setup(minor);
	return ret;
}

static int open_ne(int minor, int option) {
	if(minor >= NE_NUM) return -1;
	ne[minor].option = option;
	ne[minor].nic = NIC_UNDEF;
	if(option == -1) return 0;
	if(ne_attach(minor) == -1) return -1;
	ne_start(minor);
	return 0;
}

static int close_ne(int minor) {
	if(ne[minor].option != -1) {
		ne_outb(minor, NE_P0_IMR, 0);
		ne_stop(minor);
	}
	return 0;
}

static int write_ne(int minor, char *buffer, int size) {
	NE_ISR	status;
	int	ret;

	if(size == 0) return 0;
	if(size > NE_TX_PAGE_SIZE) return -1;
	status.BYTE = ne_inb(minor, NE_P0_ISR);
	if(status.BIT.CNT || status.BIT.OVW || status.BIT.RXE || status.BIT.TXE) {
		ne_stop(minor);
		ne_reset(minor);
		ne_setup(minor);
		ne_start(minor);
		return -1;
	}
	ret = ne_writemem(minor, buffer, ne[minor].start, size);
	if(ret == -1) {
		ne_stop(minor);
		ne_reset(minor);
		ne_setup(minor);
		ne_start(minor);
		return -1;
	}
	ret = ne_transmit(minor, size);
	if(ret == -1) ret = ne_transmit(minor, size);
	return ret;
}

static int read_ne(int minor, char *buffer, int size) {
	RSR	 rs_stat;
	NE_ISR	 int_stat;
	unsigned short bound, current, next_bound;
	unsigned short rx_start, rx_len, remain_len, sub_len;
	char	 header[4], *data;
	unsigned int value;
	int	 datasize;

	datasize = size;
	ne_outb(minor, NE_P0_COMMAND, NE_CR_STA);
	rs_stat.BYTE = ne_inb(minor, NE_P0_RSR);
	int_stat.BYTE = ne_inb(minor, NE_P0_ISR);
	if(int_stat.BIT.OVW == 1 || int_stat.BIT.CNT == 1 || int_stat.BIT.RXE || int_stat.BIT.TXE) {
		ne_stop(minor);
		ne_reset(minor);
		ne_setup(minor);
		ne_start(minor);
		ne_outb(minor, NE_P0_ISR, 0xff);
		return 0;
	}
	if(rs_stat.BIT.PRX == 0) return 0;
	bound = ne_inb(minor, NE_P0_BNRY);
	bound++;
	bound <<= 8;
	if(bound >= ne[minor].rx_end) bound = ne[minor].rx_start;
	ne_outb(minor, NE_P0_COMMAND, NE_CR_PS1);
	current = ne_inb(minor, NE_P1_CURR);
	current <<= 8;
	if(current >= ne[minor].rx_end) current = ne[minor].rx_start;
	ne_outb(minor, NE_P0_COMMAND, NE_CR_PS0);
	if(current == bound) return 0;
	ne_readmem(minor, bound, header, 4);
	rs_stat.BYTE =  header[0];
	next_bound = (unsigned short)header[1] << 8;
	rx_start = bound + 4;
	rx_len = (unsigned short)header[3] & 0xff;
	rx_len <<= 8;
	rx_len |= (unsigned short)header[2] & 0xff;
	rx_len -= 4;
	if(rx_len >= sizeof(Ether) && rx_len <= ETHER_MAX_LEN) {
		remain_len = rx_len;
		sub_len = ne[minor].rx_end - rx_start;
		data = buffer;
		if(sub_len < rx_len) {
			if(datasize < sub_len) {
				ne_readmem(minor, rx_start, data, datasize);
				datasize = 0;
			} else {
				ne_readmem(minor, rx_start, data, sub_len);
				datasize -= sub_len;
			}
			rx_start = ne[minor].rx_start;
			data = &(data[sub_len]);
			remain_len = rx_len - sub_len;
		}
		if(datasize < remain_len) {
			ne_readmem(minor, rx_start, data, datasize);
			datasize = 0;
		} else {
			ne_readmem(minor, rx_start, data, remain_len);
			datasize -= remain_len;
		}
	} else {
		rx_len = 0;
	}
	bound = (next_bound == ne[minor].rx_start) ? ne[minor].rx_end : next_bound;
	value = ((bound >> 8) & 0xff) - 1;
	ne_outb(minor, NE_P0_BNRY, value);
	ne_outb(minor, NE_P0_ISR, 0xff);
	return size - datasize;
}

static int seek_ne(int minor, int position) {
	return -1;
}

static int ioctl_ne(int minor, int data, int op) {
	NE_ISR	status;
	int	ret;

	if(minor >= NE_NUM) return -1;
	ret = 0;
	switch(op) {
	case DEV_INFO:
		break;
	case DEV_INTERVAL:
		status.BYTE = ne_inb(minor, NE_P0_ISR);
		if(status.BIT.CNT || status.BIT.OVW || status.BIT.RXE || status.BIT.TXE) {
			ne_stop(minor);
			ne_reset(minor);
			ne_setup(minor);
			ne_start(minor);
		}
		break;
	case DEV_RESET:
		ne_stop(minor);
		ne_reset(minor);
		ne_setup(minor);
		ne_start(minor);
		break;
	case GET_MAC:
		memcpy((char*)data, (char*)ne[minor].mac, 6);
		break;
	case MAXPACKET:
		ret = ETHER_MAX_LEN;
		break;
	default:
		ret = -1;
	}
	return ret;
}

static Functions func;

int device_main(int argc, char **argv) {
	unsigned int	minor, i, wait[NE_NUM];

	ne = (NEInfo*)malloc(sizeof(NEInfo) * NE_NUM);

#ifndef _KERNEL_
	for(minor = 0;minor < NE_NUM;minor++) wait[minor] = 63;
	minor = 0;
	for(i = 1;i < argc;i++) {
		if(memcmp(argv[i], "ne", 2) == 0) {
			minor = stoi(&argv[i][2]);
			if(minor >= NE_NUM) minor = 0;
		} else if(memcmp(argv[i], "b=", 2) == 0) {
			ne[minor].base = (char*)stoh(&argv[i][2]);
		} else if(memcmp(argv[i], "bw=", 3) == 0) {
			ne[minor].buswidth = stoi(&argv[i][3]);
		} else if(memcmp(argv[i], "w=", 2) == 0) {
			wait[minor] = stoi(&argv[i][2]);
		}
	}
#endif

	for(minor = 0;minor < NE_NUM;minor++) {
#ifdef _KERNEL_
		ne[minor].base = (unsigned char*)ne_config[minor].base0;
		ne[minor].buswidth = ne_config[minor].width;
		wait[minor] = ne_config[minor].wait;
#endif
		set_buswidth((int)ne[minor].base, ne[minor].buswidth, wait[minor]);
	}
	func.open_dev = open_ne;
	func.close_dev = close_ne;
	func.write_dev = write_ne;
	func.read_dev = read_ne;
	func.seek_dev = seek_ne;
	func.ioctl_dev = ioctl_ne;
	func.poll_dev = 0;
	strcpy(&(func.name[0]), "ne");
	device_return(&func);
}
