typedef volatile struct {
	int		option;
	unsigned short	start, size;
	unsigned short	rx_start, rx_end;
	unsigned char	*base;
	unsigned char	mac[6];
	unsigned char	buswidth;
	unsigned char	nic;
} NEInfo;

#define NIC_UNDEF	0
#define NIC_DP8390	1
#define NIC_AX88796B	2

#if defined(__H8300H__) || defined(__sh2__)
  typedef union {
	unsigned char BYTE;
	struct {
		unsigned char DFR :1;
		unsigned char DIS :1;
		unsigned char PHY :1;
		unsigned char MPA :1;
		unsigned char FO  :1;
		unsigned char FAE :1;
		unsigned char CRC :1;
		unsigned char PRX :1;
	} BIT;
  } RSR;
  typedef union {
	unsigned char BYTE;
	struct {
		unsigned char RST :1;
		unsigned char RDC :1;
		unsigned char CNT :1;
		unsigned char OVW :1;
		unsigned char TXE :1;
		unsigned char RXE :1;
		unsigned char PTX :1;
		unsigned char PRX :1;
	} BIT;
  } NE_ISR;
#endif

#if defined(__arm__) || defined(__sh3__)
  typedef union {
	unsigned char BYTE;
	struct {
		unsigned char PRX :1;
		unsigned char CRC :1;
		unsigned char FAE :1;
		unsigned char FO  :1;
		unsigned char MPA :1;
		unsigned char PHY :1;
		unsigned char DIS :1;
		unsigned char DFR :1;
	} BIT;
  } RSR;
  typedef union {
	unsigned char BYTE;
	struct {
		unsigned char PRX :1;
		unsigned char PTX :1;
		unsigned char RXE :1;
		unsigned char TXE :1;
		unsigned char OVW :1;
		unsigned char CNT :1;
		unsigned char RDC :1;
		unsigned char RST :1;
	} BIT;
  } NE_ISR;
#endif

#define NE_P0_COMMAND	0
#define NE_P0_PSTART	1
#define NE_P0_PSTOP	2
#define NE_P0_BNRY	3	/* Boundary Reg */
#define NE_P0_TSR	4	/* Transmit Status Reg */
#define NE_P0_TPSR	4	/* Transmit Page Start Address Reg */
#define NE_P0_NCR	5	/* Nubmer of Collisions */
#define NE_P0_TBCR0	5	/* Transmit Byte Counter Reg */
#define NE_P0_TBCR1	6
#define NE_P0_ISR	7	/* Interrupt Status Reg */
#define NE_P0_RSAR0	8	/* Remote Start Address Reg */
#define NE_P0_RSAR1	9
#define NE_P0_RBCR0	10	/* Remote Byte Count Reg */
#define NE_P0_RBCR1	11
#define NE_P0_RSR	12	/* Receive Status Reg */
#define NE_P0_RCR	12	/* Receive Configuration Reg */
#define NE_P0_TCR	13	/* Transmit Configuration Reg */
#define NE_P0_DCR	14	/* Data Configuration Reg */
#define NE_P0_IMR	15	/* Interrupt Mask Reg */
#define NE_P0_CNTR0	13
#define NE_P0_CNTR1	14
#define NE_P0_CNTR2	15
#define NE_P1_COMMAND	0
#define NE_P1_PAR0	1		/* Physical Address Reg */
#define NE_P1_PAR1	2
#define NE_P1_PAR2	3
#define NE_P1_PAR3	4
#define NE_P1_PAR4	5
#define NE_P1_PAR5	6
#define NE_P1_CURR	7		/* Current Page Reg */
#define NE_P1_MAR0	8		/* Multicast Address Reg */
#define NE_P1_MAR1	9
#define NE_P1_MAR2	10
#define NE_P1_MAR3	11
#define NE_P1_MAR4	12
#define NE_P1_MAR5	13
#define NE_P1_MAR6	14
#define NE_P1_MAR7	15
#define NE_P3_COMMAND	0
#define NE_P3_9346CR	1
#define NE_P3_BPAGE	2
#define NE_P3_CONFIG0	3
#define NE_P3_CONFIG1	4
#define NE_P3_CONFIG2	5
#define NE_P3_CONFIG3	6
#define NE_ASIX_DSR	23	/* Device Status Reg */
#define NE_ASIC_DATA	16	/* data port */
#define NE_ASIC_RESET	24	/* reset port */
#define NE_ASIC_RESET2	31	/* reset port */

#define	NE_CR_PS0	0x00	/* Page 0 Select   */
#define	NE_CR_PS1	0x40	/* Page 1 Select   */
#define	NE_CR_PS2	0x80	/* Page 2 Select   */
#define	NE_CR_PS3	0xc0	/* Page 3 Select   */
#define	NE_CR_PSMASK	0xc0	/* Page Mask       */
#define	NE_CR_RD0	0x08	/********************** 000:DON'T, 001:read  */
#define	NE_CR_RD1	0x10	/* Remote DMA Control * 010:write, 011:trans */
#define NE_CR_RD2	0x20	/********************** 1xx:disable          */
#define NE_CR_TXP	0x04	/* Transmit packet */
#define NE_CR_STA	0x02	/* start	   */
#define	NE_CR_STP	0x01	/* stop		   */

#define	NE_RCR_SEP	0x01	/* accept error packet */
#define	NE_RCR_AR	0x02	/* accept tiny packet */
#define	NE_RCR_AB	0x04	/* accept broadcast */
#define	NE_RCR_MUL	0x08	/* accept multicast */
#define	NE_RCR_PRO	0x10	/* accept all */
#define	NE_RCR_MON	0x20	/* monitor mode */

#define	NE_DCR_FT0	0x20	/* fifo  */
#define	NE_DCR_FT1	0x40	/* fifo  */
#define	NE_DCR_AR	0x10	/* Auto remote */
#define	NE_DCR_WTS	0x01	/* Word Transfer select */
#define	NE_DCR_BOS	0x02	/* Byte Transfer select */
#define	NE_DCR_LAS	0x04	/* Long Transfer select */
#define	NE_DCR_LS	0x08	/* Loopback select */

#define	NE_ISR_PRX	0x01	/* successful receive */
#define	NE_ISR_PTX	0x02	/* successful trasmit */
#define	NE_ISR_RXE	0x04	/* receive error */
#define	NE_ISR_TXE	0x08	/* transmit error */
#define	NE_ISR_OVW	0x10	/* overflow */
#define	NE_ISR_CNT	0x20	/* counter overflow */
#define	NE_ISR_RDC	0x40	/* Remote DMA complete */
#define	NE_ISR_RST	0x80	/* reset */

#define	NE_IMR_PRXE	0x01	/* successful receive */
#define	NE_IMR_PTXE	0x02	/* successful trasmit */
#define	NE_IMR_RXEE	0x04	/* receive error */
#define	NE_IMR_TXEE	0x08	/* transmit error */
#define	NE_IMR_OVWE	0x10	/* overflow */
#define	NE_IMR_CNTE	0x20	/* counter overflow */
#define	NE_IMR_RDCE	0x40	/* Remote DMA complete */

#define	NE_RSTAT_PRX	0x01	/* successful receive */
#define	NE_RSTAT_CRC	0x02	/* CRC error */
#define	NE_RSTAT_FAE	0x04	/* Frame alignment error */
#define	NE_RSTAT_OVER	0x08	/* overflow */

#define	NE_TSR_PTX	0x01	/* Transmit complete */
#define	NE_TSR_ABT	0x08	/* Transmit Aborted */

#define	NE_TCR_LB0	0x02	/* Internal loopback */
#define	NE_TCR_LB1	0x04	

#define NE_DSR_B_ENDIAN	0x80
#define NE_DSR_RDMA_RDY	0x40
#define NE_DSR_RD_RDY	0x20
#define NE_DSR_D_RDY	0x10
#define NE_DSR_I_SPEED	0x04
#define NE_DSR_I_DUPLEX	0x02
#define NE_DSR_I_LINK	0x01

#define	NE_PAGE_SIZE	 0x100
#define	NE_TX_PAGE_SIZE	 0x600
