/****************************************/
/* MES/Copyleft Yukio Mituiwa,2005	*/
/*					*/
/*  2006/6/16 first release		*/
/*					*/
/****************************************/
#include <module.h>

#ifdef _KERNEL_
  #include "mes.h"
  #include "conf.h"
  #define device_main(arg1,arg2) init_sci(void)
  extern BUSConf sci_config[];
#endif

#define FLAG_NONE		0
#define FLAG_CONTINUE	1

typedef volatile struct {
	unsigned int		freq, speed, burst;
	unsigned char		*smr;
	unsigned char		*brr;
	unsigned char		*scr;
	unsigned char		*tdr;
	unsigned char		*ssr;
	unsigned short	*ssr_w;
	unsigned char		*rdr;
	unsigned char		*fcr;
	unsigned short	*fdr;
	unsigned char		flag;
} SCIInfo;
static SCIInfo *sci;

#if defined(__H8300H__)
  #define Wait		1
  #define SCINUM	3
#endif
#if defined(__sh3__)
  #define Wait		8
  #define SCINUM	3
#endif
#if defined(__sh2__)
  #define Wait		4
  #define SCINUM	4
#endif

// SMR
#define	CA	0x80
#define	CHR	0x40
#define	PE	0x20
#define	OE	0x10
#define	STOP	0x08
#define	MP	0x04

// SCR
#define	TIE	0x80
#define	RIE	0x40
#define	TE	0x20
#define	RE	0x10
#define	MPIE	0x08
#define	TEIE	0x04

// SSR
#define	TDRE	0x80
#define	RDRF	0x40
#define	ORER	0x20
#define	FER	0x10
#define	PER	0x08
#define	TEND	0x04
#define	MPB	0x02
#define	MPBT	0x01

// SSR(FIFO)
#define	F_ER	0x80
#define	F_TEND	0x40
#define	F_TDFE	0x20
#define	F_BRK	0x10
#define	F_FER	0x08
#define	F_PER	0x04
#define	F_RDF	0x02
#define	F_DR	0x01

#define DEFAULT_BURST	10

static int ioctl_sci_freq(int, int);
static int ioctl_sci_speed(int, int);
static int open_scif(int, int);
static int write_scif(int, char*, int);
static int read_scif(int, char*, int);

#if defined(__sh3__)
static int open_scif(int minor, int option) {
	volatile int	w;

	*(sci[minor].scr) = 0x00;
	*(sci[minor].fcr) = 0x06;
	*(sci[minor].smr) = 0x00;
	for(w = 0;w < 10000;w++);
	*(sci[minor].fcr) = 0x00;
	*(sci[minor].ssr_w) = F_TEND | F_TDFE | F_RDF;
	*(sci[minor].scr) |= RE | TE;
	return 0;
}

static int write_scif(int minor, char *buffer, int size) {
	int	i;

	for(i = 0;i < size;i++) {
		while((*(sci[minor].ssr_w) & F_TDFE) == 0);
		*(sci[minor].tdr) = buffer[i];
		*(sci[minor].ssr_w) &= ~(F_TEND + F_TDFE);
	}
	return size;
}

static int read_scif(int minor, char *buffer, int size) {
	int	i;

	for(i = 0;i < size;i++) {
		if((*(sci[minor].ssr_w) & F_RDF) == 0) break;
		buffer[i]  = *(sci[minor].rdr);
		*(sci[minor].ssr_w) &= ~F_RDF;
	}
	return i;
}
#endif

static int open_sci(int minor, int option) {
	volatile int	w;

	if(minor >= SCINUM) return -1;
#if defined(__sh3__)
	if(minor == 2) {
		open_scif(minor, option);
	} else {
#endif
	*(sci[minor].smr) = 0;
	*(sci[minor].scr) = 0;
	for(w = 0;w < 3000;w++);
	*(sci[minor].scr) = RE | TE;
	w = *(sci[minor].ssr);
	*(sci[minor].ssr) = TDRE;
#if defined(__sh3__)
	}
#endif
	sci[minor].burst = DEFAULT_BURST;
	sci[minor].flag = FLAG_NONE;
	ioctl_sci_speed(minor, sci[minor].speed);
	return 0;
}

static int close_sci(int minor) {
	if(minor >= SCINUM) return -1;
	*(sci[minor].scr) = 0;
	return 0;
}
static int write_sci(int minor, char *buffer, int size) {
	int	i;

	if(minor >= SCINUM) return -1;
#if defined(__sh3__)
	if(minor == 2) return write_scif(minor, buffer, size);
#endif
	if(*(sci[minor].ssr) & (ORER | FER | PER)) {
		*(sci[minor].ssr) &= ~(ORER | FER | PER);
		return 0;
	}
	for(i = 0;i < size;i++) {
		while(!(*(sci[minor].ssr) & TDRE)) {
			if(*(sci[minor].ssr) & (ORER | FER | PER)) {
				*(sci[minor].ssr) &= ~(ORER | FER | PER);
				return 0;
			}
		}
		*(sci[minor].tdr) = buffer[i];
		*(sci[minor].ssr) &= ~TDRE;
	}
	while(!(*(sci[minor].ssr) & TEND)) {
		if(*(sci[minor].ssr) & (ORER | FER | PER)) {
			*(sci[minor].ssr) &= ~(ORER | FER | PER);
			return 0;
		}
	}
	return size;
}

static int read_sci(int minor, char *buffer, int size) {
	int	i, r;

	if(minor >= SCINUM) return -1;
#if defined(__sh3__)
	if(minor == 2) return read_scif(minor, buffer, size);
#endif
	if((sci[minor].flag & SCI_CONTINUE) == 0) {
		if(*(sci[minor].ssr) & (ORER | FER | PER)) {
			*(sci[minor].ssr) &= ~(ORER | FER | PER);
			return 0;
		}
	}
	for(i = 0;i < size;i++) {
		for(r = 0;r < (sci[minor].burst * Wait);r++) {
			if(*(sci[minor].ssr) & RDRF) break;
		}
		if(r == (sci[minor].burst * Wait)) break;
		buffer[i]  = *(sci[minor].rdr);
		*(sci[minor].ssr) &= ~RDRF;
	}
	return i;
}

static int seek_sci(int minor, int pos) {
	if(minor >= SCINUM) return -1;
	return 0;
}

static int ioctl_sci_freq(int minor, int freq) {
	if(minor >= SCINUM) return -1;
	sci[minor].freq = freq;
	return 0;
}

static int ioctl_sci_speed(int minor, int speed) {
	unsigned int	i, x, y, b, s;

	if(minor >= SCINUM) return -1;
	s = 32;
	for(i = 0;i < 4;i++) {
		x = sci[minor].freq * 10;
		y = s * speed;
		b = x / y + 5;
		b = b / 10 - 1;
		if(b <= 255) break;
		s <<= 2;
	}
	if(i == 4) return -1;
	*(sci[minor].smr) &= ~0x03;
	*(sci[minor].smr) |= i;
	*(sci[minor].brr) = b;
	return 0;
}

static int ioctl_sci(int minor, int data, int op) {
	int	i, ret;

	if(minor >= SCINUM) return -1;
	switch(op) {
	case DEV_INFO:
		ret = 0;
		break;
	case SCI_FREQ:
		ret = ioctl_sci_freq(minor, data);
		break;
	case SCI_SPEED:
		ret = ioctl_sci_speed(minor, data);
		break;
	case SCI_CONTINUE:
		ret = 0;
		if(data == 0) {
			sci[minor].flag &= ~FLAG_CONTINUE;
		} else {
			sci[minor].flag |= FLAG_CONTINUE;
		}
		break;
	case READ_BURST:
		ret = 0;
		sci[minor].burst = (data == 0) ? DEFAULT_BURST : data;
		break;
	default:
		ret = -1;
	}
	return ret;
}

static Functions func;

int device_main(int argc, char **argv) {
	unsigned int	minor;
	unsigned int	base, n;

	sci = (SCIInfo*)malloc(sizeof(SCIInfo) * SCINUM);

#ifndef _KERNEL_
	minor = 0;
	for(n = 1;n < argc;n++) {
		if(memcmp(argv[n], "sci", 3) == 0) {
			minor = stoi(&argv[n][3]);
			if(minor >= SCINUM) minor = 0;
		} else if(memcmp(argv[n], "f=", 2) == 0) {
			sci[minor].freq = stoi(&argv[n][2]);
		} else if(memcmp(argv[n], "br=", 3) == 0) {
			sci[minor].speed = stoi(&argv[n][3]);
		}
	}
#endif

	for(minor = 0;minor < SCINUM;minor++) {

#ifdef _KERNEL_
		sci[minor].freq = sci_config[minor].scifreq;
		sci[minor].speed = sci_config[minor].scispeed;
#endif
#if defined(__sh3__)
		if(minor == 2) {
			base = 0xa4000150;
			sci[minor].smr = (unsigned char*)(base + 0);
			sci[minor].brr = (unsigned char*)(base + 2); 
			sci[minor].scr = (unsigned char*)(base + 4);
			sci[minor].tdr = (unsigned char*)(base + 6);
			sci[minor].ssr_w = (unsigned short*)(base + 8);
			sci[minor].rdr = (unsigned char*)(base + 10);
			sci[minor].fcr = (unsigned char*)(base + 12);
			sci[minor].fdr = (unsigned short*)(base + 14);
		} else {
			base = 0xfffffe80 * minor * 0x10;
			sci[minor].smr = (unsigned char*)(base + 0);
			sci[minor].brr = (unsigned char*)(base + 2); 
			sci[minor].scr = (unsigned char*)(base + 4);
			sci[minor].tdr = (unsigned char*)(base + 6);
			sci[minor].ssr = (unsigned char*)(base + 8);
			sci[minor].rdr = (unsigned char*)(base + 10);
		}
#endif
#if defined(__sh2__)
		base = 0xffff81a0 + minor * 16;
#endif
#if defined(__H8300H__)
		base = 0xffffb0 + minor * 8;
#endif
#if defined(__H8300H__) || defined(__sh2__)
		sci[minor].smr = (unsigned char*)(base + 0);
		sci[minor].brr = (unsigned char*)(base + 1); 
		sci[minor].scr = (unsigned char*)(base + 2);
		sci[minor].tdr = (unsigned char*)(base + 3);
		sci[minor].ssr = (unsigned char*)(base + 4);
		sci[minor].rdr = (unsigned char*)(base + 5);
#endif
	}
	func.open_dev = open_sci;
	func.close_dev = close_sci;
	func.write_dev = write_sci;
	func.read_dev = read_sci;
	func.seek_dev = seek_sci;
	func.ioctl_dev = ioctl_sci;
	func.poll_dev = 0;
	strcpy(&(func.name[0]), "com");
	device_return(&func);
}
